# -*- coding: utf-8 -*-
"""Implements a class to test the test profile worker.
"""
from tlsmate.workers.server_profile import ReadProfileWorker, DumpProfileWorker
from tlsmate.config import Configuration
from tlsmate.tlsmate import TlsMate
from tlsmate.structs import ConfigItem


def test_server_profile_read_dump(server_profile, capsys):
    with open(server_profile) as fd:
        server_file = fd.read()

    config = Configuration()
    config.register(ConfigItem("write_profile"))
    config.register(ConfigItem("read_profile"))
    config.register(ConfigItem("format", type=str, default="yaml"))
    config.set("read_profile", str(server_profile))
    tlsmate = TlsMate(config)
    ReadProfileWorker(tlsmate).run()
    DumpProfileWorker(tlsmate).run()

    captured = capsys.readouterr()
    a1 = server_file.split("\n")
    a2 = captured.out.split("\n")
    for a, b in zip(a1, a2):
        assert a == b



# -*- coding: utf-8 -*-
"""Implement unit tests for the module recorder.
"""
import datetime
import time
from tlsmate.recorder import SocketEvent
import pytest
import requests


class Response(object):
    pass


def test_inject_basic(tlsmate):

    recorder = tlsmate.recorder
    recorder.trace(ec_seed=1)
    assert recorder.is_injecting() is False
    assert recorder.is_recording() is False

    recorder.record()
    assert recorder.is_injecting() is False
    assert recorder.is_recording() is True

    data = bytes.fromhex("00 11 22 33 44 55 66 77 88 99 aa bb cc")
    assert recorder.inject(verify_data_finished_sent=data) == data
    assert recorder.inject(ec_seed=5) == 5
    assert recorder.inject(openssl_command="hello world")
    assert recorder.inject(timestamp=15.9)
    assert recorder.inject(client_auth=True)

    recorder.replay()
    assert recorder.is_injecting() is True
    assert recorder.is_recording() is False

    assert recorder.inject(verify_data_finished_sent=b"deadbeef") == data
    assert recorder.inject(ec_seed=10) == 5
    assert recorder.inject(openssl_command="baeh!") == "hello world"
    assert recorder.inject(timestamp=15000.9) == 15.9
    assert recorder.inject(client_auth=False) is True

    recorder.deactivate()
    assert recorder.is_injecting() is False
    assert recorder.is_recording() is False


def test_client_auth(tlsmate):
    recorder = tlsmate.recorder
    recorder.record()
    recorder.trace_client_auth(("hello", "world"))

    recorder.replay()
    assert recorder.get_client_auth() == [("hello", "world")]


def test_socket_recv(tlsmate):
    recorder = tlsmate.recorder
    recorder.record()
    assert recorder.inject_socket_recv() is None
    data = bytes.fromhex("00 11 22 33 44 55 66 77 88 99 aa bb cc")
    recorder.additional_delay(0.500)
    recorder.additional_delay(0.200)
    recorder.trace_socket_recv(0.050, SocketEvent.DATA, data)

    recorder.replay()
    start = datetime.datetime.now()
    assert recorder.inject_socket_recv() == data
    duration = datetime.datetime.now() - start
    assert 700 <= duration.microseconds / 1000 <= 800


def test_reponse(tlsmate):
    recorder = tlsmate.recorder
    recorder.record()
    data = bytes.fromhex("00 11 22 33 44 55 66 77 88 99 aa bb cc")
    resp = Response()
    resp.ok = True
    resp.content = data
    resp.status_code = 200
    recorder.trace_response(0.050, SocketEvent.DATA, resp)

    recorder.replay()
    resp = recorder.inject_response()
    assert resp.ok is True
    assert resp.content == data
    assert resp.status_code == 200


def test_timestamp(tlsmate):
    recorder = tlsmate.recorder
    recorder.record()
    timestamp1 = recorder.get_timestamp()
    time.sleep(0.1)
    recorder.replay()
    timestamp2 = recorder.get_timestamp()
    assert timestamp1 == timestamp2


def test_send_all(tlsmate):
    recorder = tlsmate.recorder
    recorder.record()
    data = bytes.fromhex("00 11 22 33 44 55 66 77 88 99 aa bb cc")
    assert recorder.trace_socket_sendall(data) is True


def test_inject_resonse_closure(tlsmate):
    recorder = tlsmate.recorder
    recorder.record()
    recorder.trace_response(0.010, SocketEvent.CLOSURE)
    recorder.replay()
    with pytest.raises(Exception):
        recorder.inject_response()


def test_inject_resonse_timeout(tlsmate):
    recorder = tlsmate.recorder
    recorder.record()
    assert recorder.inject_response() is None
    recorder.trace_response(0.010, SocketEvent.TIMEOUT)
    recorder.replay()
    with pytest.raises(requests.Timeout):
        recorder.inject_response()



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
from tests.cipher_suite_tester import CipherSuiteTester
from tlsmate import tls
from tlsmate import msg
from tlsmate.tlssuite import TlsLibrary


class TestCase(CipherSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the CipherSuiteTester class.
    """

    name = "ClientProfiles"
    path = pathlib.Path(__file__)
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-rsa --cert2 server-ecdsa -- -www"
    )
    library = TlsLibrary.openssl1_1_1

    # Uncomment the line below if you do not want to use the default version and
    # adapt it to your needs.
    # version = tls.Version.TLS12

    def run(self, tlsmate, is_replaying=False):
        client = tlsmate.client

        client.set_profile(tls.Profile.LEGACY)
        end_of_tc_reached = False
        with client.create_connection() as conn:
            conn.send(msg.ClientHello)
            conn.wait(msg.ServerHello)
            end_of_tc_reached = True
        assert end_of_tc_reached is True

        client.set_profile(tls.Profile.INTEROPERABILITY)
        end_of_tc_reached = False
        with client.create_connection() as conn:
            conn.send(msg.ClientHello)
            conn.wait(msg.ServerHello)
            end_of_tc_reached = True
        assert end_of_tc_reached is True

        client.set_profile(tls.Profile.MODERN)
        end_of_tc_reached = False
        with client.create_connection() as conn:
            conn.send(msg.ClientHello)
            conn.wait(msg.ServerHello)
            end_of_tc_reached = True
        assert end_of_tc_reached is True

        client.set_profile(tls.Profile.TLS13)
        end_of_tc_reached = False
        with client.create_connection() as conn:
            conn.send(msg.ClientHello)
            conn.wait(msg.ServerHello)
            end_of_tc_reached = True
        assert end_of_tc_reached is True


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Module scanning for renegotiation support
"""
# import basic stuff

# import own stuff
import tlsmate.plugin as plg
import tlsmate.tls as tls

# import other stuff


class ScanRenegotiation(plg.Worker):
    name = "renegotiation"
    descr = "scan for renegotiation support"
    prio = 30

    def run(self):

        self.server_profile.allocate_features()
        versions = [tls.Version.TLS10, tls.Version.TLS11, tls.Version.TLS12]
        prof_values = self.server_profile.get_profile_values(versions, full_hs=True)
        if not prof_values.versions:
            self.server_profile.features.insecure_renegotiation = tls.ScanState.NA
            self.server_profile.features.secure_renegotation = tls.ScanState.NA
            self.server_profile.features.scsv_renegotiation = tls.ScanState.NA
            return

        self.server_profile.features.insecure_renegotiation = tls.ScanState.UNDETERMINED
        self.server_profile.features.secure_renegotation = tls.ScanState.UNDETERMINED
        self.server_profile.features.scsv_renegotiation = tls.ScanState.UNDETERMINED
        self.client.init_profile(profile_values=prof_values)
        self.server_profile.features.insecure_renegotiation = tls.ScanState.FALSE
        with self.client.create_connection() as conn:
            conn.handshake()
            conn.handshake()
            if conn.handshake_completed:
                self.server_profile.features.insecure_renegotiation = tls.ScanState.TRUE

        self.server_profile.features.secure_renegotation = tls.ScanState.FALSE
        self.client.profile.support_secure_renegotiation = True
        with self.client.create_connection() as conn:
            conn.handshake()
            conn.handshake()
            if conn.handshake_completed:
                self.server_profile.features.secure_renegotation = tls.ScanState.TRUE

        self.server_profile.features.scsv_renegotiation = tls.ScanState.FALSE
        self.client.profile.support_secure_renegotiation = False
        self.client.profile.support_scsv_renegotiation = True
        with self.client.create_connection() as conn:
            conn.handshake()
            conn.handshake()
            if conn.handshake_completed:
                self.server_profile.features.scsv_renegotiation = tls.ScanState.TRUE



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
from tlsmate.workers.eval_cipher_suites import ScanCipherSuites
from tlsmate.workers.supported_groups import ScanSupportedGroups
from tlsmate.workers.sig_algo import ScanSigAlgs
from tlsmate.tlssuite import TlsSuiteTester, TlsLibrary


class TestCase(TlsSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the TcRecorder class.
    """

    sp_out_yaml = "profile_basic_wolfssl3_12_0"
    recorder_yaml = "recorder_basic_profile_wolfssl3_12_0"
    path = pathlib.Path(__file__)
    server_cmd = "utils/start_wolfssl --version {library} --port {server_port}"
    library = TlsLibrary.wolfssl3_12_0

    server = "localhost"

    def run(self, tlsmate, is_replaying):
        for vers in ["sslv2", "sslv3", "tls10", "tls11", "tls12", "tls13"]:
            tlsmate.config.set(vers, True)
        ScanCipherSuites(tlsmate).run()
        ScanSupportedGroups(tlsmate).run()
        ScanSigAlgs(tlsmate).run()


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Module for downgrade attack prevention worker
"""
# import basic stuff

# import own stuff
import tlsmate.msg as msg
import tlsmate.plugin as plg
import tlsmate.tls as tls

# import other stuff


class ScanDowngrade(plg.Worker):
    name = "downgrade"
    descr = "scan for downgrade attack prevention"
    prio = 30

    def run(self):
        versions = self.server_profile.get_versions()
        nbr_versions = len(versions)
        if tls.Version.SSL20 in versions:
            nbr_versions -= 1

        if nbr_versions < 2:
            status = tls.ScanState.NA

        else:
            prof_values = self.server_profile.get_profile_values([versions[-2]])
            self.client.init_profile(profile_values=prof_values)
            self.client.profile.cipher_suites.append(tls.CipherSuite.TLS_FALLBACK_SCSV)

            status = tls.ScanState.UNDETERMINED
            with self.client.create_connection() as conn:
                conn.send(msg.ClientHello)
                response = conn.wait(msg.Any)
                if isinstance(response, msg.ServerHello):
                    status = tls.ScanState.FALSE

                elif isinstance(response, msg.Alert):
                    status = tls.ScanState.TRUE

        self.server_profile.allocate_features()
        self.server_profile.features.downgrade_attack_prevention = status



# -*- coding: utf-8 -*-
"""Module for OCSP stapling worker
"""
# import basic stuff

# import own stuff
import tlsmate.plugin as plg
import tlsmate.tls as tls

# import other stuff


class ScanOcspStapling(plg.Worker):
    name = "ocsp_stapling"
    descr = "scan for OCSP stapling"
    prio = 30

    def _scan_stapling(self):

        versions = tls.Version.tls_only()
        prof_values = self.server_profile.get_profile_values(versions, full_hs=True)
        if not prof_values.versions:
            status = tls.ScanState.NA

        else:
            self.client.init_profile(profile_values=prof_values)
            self.client.alert_on_invalid_cert = False
            self.client.profile.support_status_request = True
            with self.client.create_connection() as conn:
                conn.handshake()

            # Hm, some TLS libraries return the status_request(_v2) extension in the
            # server_hello, and some are not (e.g. openssl for TLS13). So check if
            # OCSP status request is received without evaluating it. Evaluating
            # the responses would require more sophisticated handling, as the state
            # should then be determined for all certificate chains, respectively
            # for all certificates (status_request_v2) for which a response is
            # received.

            if not conn.handshake_completed:
                status = tls.ScanState.UNDETERMINED

            elif conn.stapling_status:
                status = tls.ScanState.TRUE

            else:
                status = tls.ScanState.FALSE

        self.server_profile.features.ocsp_stapling = status

    def _scan_multi_stapling(self):

        prof_values = self.server_profile.get_profile_values(
            [tls.Version.TLS10, tls.Version.TLS11, tls.Version.TLS12], full_hs=True
        )
        if not prof_values.versions:
            status = tls.ScanState.NA

        else:
            self.client.init_profile(profile_values=prof_values)
            self.client.alert_on_invalid_cert = False
            self.client.profile.support_status_request_v2 = tls.StatusType.OCSP_MULTI
            with self.client.create_connection() as conn:
                conn.handshake()

            if conn.msg.server_hello is None:
                status = tls.ScanState.UNDETERMINED

            elif conn.msg.server_hello.get_extension(tls.Extension.STATUS_REQUEST_V2):
                status = tls.ScanState.TRUE

            else:
                status = tls.ScanState.FALSE

        self.server_profile.features.ocsp_multi_stapling = status

    def run(self):
        self.server_profile.allocate_features()
        self._scan_stapling()
        self._scan_multi_stapling()



# -*- coding: utf-8 -*-
"""Module for the scan plugin
"""
# import basic stuff
from pathlib import Path
import argparse
from typing import Any

# import own stuff
import tlsmate.config as conf
import tlsmate.plugin as plg
import tlsmate.structs as structs
import tlsmate.utils as utils
import tlsmate.workers.base_vulnerabilities as base_vulnerabilities
import tlsmate.workers.ccs_injection as ccs_injection
import tlsmate.workers.compression as compression
import tlsmate.workers.dh_params as dh_params
import tlsmate.workers.downgrade as downgrade
import tlsmate.workers.encrypt_then_mac as encrypt_then_mac
import tlsmate.workers.ephemeral_key_reuse as ephemeral_key_reuse
import tlsmate.workers.eval_cipher_suites as eval_cipher_suites
import tlsmate.workers.grease as grease
import tlsmate.workers.heartbeat as heartbeat
import tlsmate.workers.heartbleed as heartbleed
import tlsmate.workers.master_secret as master_secret
import tlsmate.workers.ocsp_stapling as ocsp_stapling
import tlsmate.workers.padding_oracle as padding_oracle
import tlsmate.workers.renegotiation as renegotiation
import tlsmate.workers.resumption as resumption
import tlsmate.workers.robot as robot
import tlsmate.workers.scanner_info as scanner_info
import tlsmate.workers.server_profile as server_profile
import tlsmate.workers.sig_algo as sig_algo
import tlsmate.workers.supported_groups as supported_groups
import tlsmate.workers.text_server_profile as text_server_profile


class ArgPort(plg.Plugin):
    """Argument for port.
    """

    config = conf.config_port
    cli_args = plg.Args(
        "--port",
        default=None,
        help="the port number of the host [0-65535]. Defaults to 443.",
        type=int,
    )

    @classmethod
    def args_parsed(
        cls,
        args: Any,
        parser: argparse.ArgumentParser,
        subcommand: str,
        config: conf.Configuration,
    ) -> None:
        if args.port is not None and (args.port < 0 or args.port > 0xFFFF):
            parser.error("port must be in the range [0-65535]")
        super().args_parsed(args, parser, subcommand, config)


class ArgInterval(plg.Plugin):
    """Argument for interval.
    """

    config = conf.config_interval
    cli_args = plg.Args(
        "--interval",
        default=0,
        help="the interval in milliseconds between two handshakes.",
        type=int,
    )


class ArgIpv6Preference(plg.Plugin):
    """Argument for IPv6 preference.
    """

    config = conf.config_ipv6_preference
    cli_args = plg.Args(
        "--ipv6-preference",
        help="prefers resolved IPv6 addresses over IPv4. Defaults to False.",
        action=utils.BooleanOptionalAction,
    )


class ArgKeyLogFile(plg.Plugin):
    """Argument for key logging file.
    """

    config = conf.config_key_log_file
    cli_args = plg.Args(
        "--key-log-file",
        default=None,
        help=(
            "write to a key log file which can be used by wireshark to decode "
            "encrypted traffic."
        ),
    )


class ArgProgress(plg.Plugin):
    """Argument for progress.
    """

    config = conf.config_progress
    cli_args = plg.Args(
        "--progress",
        help="provides a progress indicator. Defaults to False.",
        action=utils.BooleanOptionalAction,
    )


class ArgProxy(plg.Plugin):
    """Argument for proxy.
    """

    config = conf.config_proxy
    cli_args = plg.Args(
        "--proxy",
        type=str,
        help=(
            "the URL of the proxy. Must include the scheme and may include the user, "
            "password and port, e.g.: `http://user:password@myproxy.net:3128`."
        ),
    )


class ArgSni(plg.Plugin):
    """Argument for sni.
    """

    config = conf.config_sni
    cli_args = plg.Args(
        "--sni",
        type=str,
        help=(
            "the server name indication, i.e., the domain name of the server to "
            "contact. If not given, the value will be taken from the host "
            "parameter. This parameter is useful, if the host is given as an "
            "IP address."
        ),
    )


class ArgHost(plg.Plugin):
    """Argument for host.
    """

    config = conf.config_host
    cli_args = plg.Args(
        "host",
        help=(
            "the target host. Can be given as a domain name or as an IPv4/IPv6 address."
        ),
        type=str,
    )


class GroupBasicScan(plg.Plugin):
    """Group for basic arguments.
    """

    plugins = [
        ArgPort,
        ArgInterval,
        ArgIpv6Preference,
        ArgKeyLogFile,
        ArgProgress,
        ArgProxy,
        ArgSni,
        ArgHost,
    ]


class ArgReadProfile(plg.Plugin):
    """Argument for read-profile.
    """

    config = structs.ConfigItem("read_profile", type=str, default=None)
    cli_args = plg.Args(
        "--read-profile",
        type=str,
        help=(
            "JSON/Yaml file to read the server profile from. The format will be "
            "determined automatically."
        ),
    )
    workers = [server_profile.ReadProfileWorker]


class ArgWriteProfile(plg.Plugin):
    """Argument for write-profile.
    """

    config = structs.ConfigItem("write_profile", type=str, default=None)
    cli_args = plg.Args(
        "--write-profile",
        type=str,
        help=(
            "writes the server profile to the given file. If no file is given, "
            'the profile will be dumped to STDOUT (unless "--format=none" is '
            "given)."
        ),
    )


class ArgFormat(plg.Plugin):
    """Argument for format.
    """

    config = structs.ConfigItem("format", type=str, default="text")
    cli_args = plg.Args(
        "--format",
        choices=["text", "html", "json", "yaml", "none"],
        help=('the output format of the server profile. Defaults to "text".'),
        default=None,
    )

    @classmethod
    def args_parsed(cls, args, parser, subcommand, config):
        super().args_parsed(args, parser, subcommand, config)
        format_type = config.get("format")
        if format_type in ["text", "html"]:
            plg.WorkManager.register(text_server_profile.TextProfileWorker)
        elif format_type in ["json", "yaml"]:
            plg.WorkManager.register(server_profile.DumpProfileWorker)


class ArgColor(plg.Plugin):
    """Argument for color.
    """

    config = structs.ConfigItem("color", type=bool, default=True)
    cli_args = plg.Args(
        "--color",
        help=(
            "use colored console output. Only used if --format=text is given. "
            "Defaults to True."
        ),
        action=utils.BooleanOptionalAction,
    )


class ArgStyle(plg.Plugin):
    """Argument for style.
    """

    _DEFAULT_STYLE = Path(__file__).parent / "../styles/default.yaml"
    config = structs.ConfigItem(
        "style", type=str, default=str(_DEFAULT_STYLE.resolve())
    )
    cli_args = plg.Args(
        "--style",
        type=str,
        help=(
            "a yaml file defining the text output and the color scheme used if "
            "--format=text or --format=html is given. If not given, the internal "
            "default file will be used."
        ),
    )


class GroupServerProfile(plg.Plugin):
    """Argument group for server profile arguments.
    """

    group = plg.Args(title="Server profile options")
    plugins = [
        ArgWriteProfile,
        ArgFormat,
        ArgColor,
        ArgStyle,
    ]


class ArgCaCerts(plg.Plugin):
    """Argument for CA certs.
    """

    config = conf.config_ca_certs
    cli_args = plg.Args(
        "--ca-certs",
        nargs="*",
        type=str,
        help=(
            "list of root-ca certificate files. Each file may contain multiple "
            "root-CA certificates in PEM format. Certificate chains received from "
            "the server will be validated against this set of root certificates."
        ),
    )


class ArgClientKey(plg.Plugin):
    """Argument for client keys.
    """

    config = conf.config_client_key
    cli_args = plg.Args(
        "--client-key",
        type=str,
        nargs="*",
        help=(
            "a list of files containing the client private keys in PEM format. "
            "Used for client authentication."
        ),
        default=None,
    )


class ArgClientChain(plg.Plugin):
    """Argument for client certificate chains.
    """

    config = conf.config_client_chain
    cli_args = plg.Args(
        "--client-chain",
        type=str,
        nargs="*",
        help=(
            "a list of files containing the certificate chain used for client "
            "authentication in PEM format. The number of given files must be the "
            "same than the number of given client key files. This first given "
            "chain file corresponds to the first given client key file, and so on."
        ),
    )


class ArgCrl(plg.Plugin):
    """Argument for CRL.
    """

    config = conf.config_crl
    cli_args = plg.Args(
        "--crl",
        help=(
            "download the CRL to check for the certificate revocation status. "
            "Defaults to True."
        ),
        action=utils.BooleanOptionalAction,
    )


class ArgOcsp(plg.Plugin):
    """Argument for OCSP.
    """

    config = conf.config_ocsp
    cli_args = plg.Args(
        "--ocsp",
        help=(
            "query the OCSP servers for checking the certificate "
            "revocation status. Defaults to True."
        ),
        action=utils.BooleanOptionalAction,
    )


class GroupX509(plg.Plugin):
    """Argument group for X509 arguments.
    """

    group = plg.Args(title="X509 certificates options")
    plugins = [ArgCaCerts, ArgClientKey, ArgClientChain, ArgCrl, ArgOcsp]

    @classmethod
    def args_parsed(cls, args, parser, subcommand, config):
        if (args.client_key is not None) or (args.client_chain is not None):
            if (args.client_key is None) or (args.client_chain is None):
                parser.error(
                    "if --client-key is given, --client-chain must be given as well, "
                    "and vice versa"
                )
            if len(args.client_key) != len(args.client_chain):
                parser.error(
                    "number of arguments for --client-key and --client-chain must "
                    "be identical"
                )

        super().args_parsed(args, parser, subcommand, config)


class ArgSslv2(plg.Plugin):
    """Argument for SSLv2.
    """

    config = structs.ConfigItem("sslv2", type=bool, default=None)
    cli_args = plg.Args(
        "--sslv2",
        help="scan for protocol version SSLv2",
        action=utils.BooleanOptionalAction,
    )


class ArgSslv3(plg.Plugin):
    """Argument for SSLv3.
    """

    config = structs.ConfigItem("sslv3", type=bool, default=None)
    cli_args = plg.Args(
        "--sslv3",
        help="scan for protocol version SSLv3",
        action=utils.BooleanOptionalAction,
    )


class ArgTls10(plg.Plugin):
    """Argument for TLS1.0.
    """

    config = structs.ConfigItem("tls10", type=bool, default=None)
    cli_args = plg.Args(
        "--tls10",
        help="scan for protocol version TLS1.0",
        action=utils.BooleanOptionalAction,
    )


class ArgTls11(plg.Plugin):
    """Argument for TLS1.1.
    """

    config = structs.ConfigItem("tls11", type=bool, default=None)
    cli_args = plg.Args(
        "--tls11",
        help="scan for protocol version TLS1.1",
        action=utils.BooleanOptionalAction,
    )


class ArgTls12(plg.Plugin):
    """Argument for TLS1.2.
    """

    config = structs.ConfigItem("tls12", type=bool, default=None)
    cli_args = plg.Args(
        "--tls12",
        help="scan for protocol version TLS1.2",
        action=utils.BooleanOptionalAction,
    )


class ArgTls13(plg.Plugin):
    """Argument for TLS1.3.
    """

    config = structs.ConfigItem("tls13", type=bool, default=None)
    cli_args = plg.Args(
        "--tls13",
        help="scan for protocol version TLS1.3",
        action=utils.BooleanOptionalAction,
    )


class GroupTlsVersions(plg.Plugin):
    """Argument group for TLS versions.
    """

    group = plg.Args(
        title="TLS protocol versions",
        description=(
            "The following options specify the TLS protocol versions to scan. "
            "By default all versions will be scanned, but if one version is "
            "explicitly set to true, all other versions will be defaulted to false."
        ),
    )
    plugins = [ArgSslv2, ArgSslv3, ArgTls10, ArgTls11, ArgTls12, ArgTls13]

    @classmethod
    def args_parsed(cls, args, parser, subcommand, config):
        super().args_parsed(args, parser, subcommand, config)
        default = all([config.get(vers.config.name) is None for vers in cls.plugins])
        enabled = False
        for vers in cls.plugins:
            val = config.get(vers.config.name)
            if val is None:
                val = default
                config.set(vers.config.name, default)
            enabled = enabled or val

        if not enabled:
            parser.error("at least one TLS version must be given")


class ArgFeatures(plg.Plugin):
    """Argument for Features.
    """

    config = structs.ConfigItem("features", type=bool, default=True)
    cli_args = plg.Args(
        "--features",
        help=(
            "specifies whether to include or exclude all features in the scan. "
            "Per feature this behavior can be overruled by its specific command "
            "line option below. "
            "Defaults to true if no specific feature is enabled, otherwise it "
            "defaults to false."
        ),
        action=utils.BooleanOptionalAction,
    )


class ArgCompression(plg.Plugin):
    """Argument for compression feature.
    """

    config = structs.ConfigItem("compression", type=bool, default=None)
    cli_args = plg.Args(
        "--compression",
        help="scan for compression support",
        action=utils.BooleanOptionalAction,
    )
    workers = [compression.ScanCompression]


class ArgDhGroup(plg.Plugin):
    """Argument for DH groups.
    """

    config = structs.ConfigItem("dh_groups", type=bool, default=None)
    cli_args = plg.Args(
        "--dh-groups",
        help="scan for finite field DH groups (only TL1.0 - TLS1.2)",
        action=utils.BooleanOptionalAction,
    )
    workers = [dh_params.ScanDhGroups]


class ArgEncThenMac(plg.Plugin):
    """Argument for encrypt-then-mac extension.
    """

    config = structs.ConfigItem("encrypt_then_mac", type=bool, default=None)
    cli_args = plg.Args(
        "--encrypt-then-mac",
        help="scan for encrypt-then-mac support (only TL1.0 - TLS1.2)",
        action=utils.BooleanOptionalAction,
    )
    workers = [encrypt_then_mac.ScanEncryptThenMac]


class ArgEphemKeyReuse(plg.Plugin):
    """Argument for ephemeral key reuse.
    """

    config = structs.ConfigItem("ephemeral_key_reuse", type=bool, default=None)
    cli_args = plg.Args(
        "--ephemeral-key-reuse",
        help="scan for reuse of ephemeral keys",
        action=utils.BooleanOptionalAction,
    )
    workers = [ephemeral_key_reuse.ScanEphemeralKeyReuse]


class ArgExtMasterSecret(plg.Plugin):
    """Argument for extended-master-secret extension.
    """

    config = structs.ConfigItem("ext_master_secret", type=bool, default=None)
    cli_args = plg.Args(
        "--ext-master-secret",
        help="scan for extended master secret support (only TL1.0 - TLS1.2)",
        action=utils.BooleanOptionalAction,
    )
    workers = [master_secret.ScanExtendedMasterSecret]


class ArgFallback(plg.Plugin):
    """Argument for fallback.
    """

    config = structs.ConfigItem("fallback", type=bool, default=None)
    cli_args = plg.Args(
        "--fallback",
        help="scan for downgrade attack prevention (TLS_FALLBACK_SCSV)",
        action=utils.BooleanOptionalAction,
    )
    workers = [downgrade.ScanDowngrade]


class ArgGrease(plg.Plugin):
    """Argument for GREASE.
    """

    config = structs.ConfigItem("grease", type=bool, default=None)
    cli_args = plg.Args(
        "--grease",
        help="scan for unknown parameter tolerance",
        action=utils.BooleanOptionalAction,
    )
    workers = [grease.ScanGrease]


class ArgHeartbeat(plg.Plugin):
    """Argument for heartbeat extension.
    """

    config = structs.ConfigItem("heartbeat", type=bool, default=None)
    cli_args = plg.Args(
        "--heartbeat",
        help="scan for heartbeat support",
        action=utils.BooleanOptionalAction,
    )
    workers = [heartbeat.ScanHeartbeat]


class ArgOcspStapling(plg.Plugin):
    """Argument for OCSP stapling.
    """

    config = structs.ConfigItem("ocsp_stapling", type=bool, default=None)
    cli_args = plg.Args(
        "--ocsp-stapling",
        help="scan for OCSP stapling support",
        action=utils.BooleanOptionalAction,
    )
    workers = [ocsp_stapling.ScanOcspStapling]


class ArgRenegotiation(plg.Plugin):
    """Argument for renegotiation feature.
    """

    config = structs.ConfigItem("renegotiation", type=bool, default=None)
    cli_args = plg.Args(
        "--renegotiation",
        help="scan for renegotiation support (SSL30 - TLS1.2)",
        action=utils.BooleanOptionalAction,
    )
    workers = [renegotiation.ScanRenegotiation]


class ArgResumption(plg.Plugin):
    """Argument for resumption feature.
    """

    config = structs.ConfigItem("resumption", type=bool, default=None)
    cli_args = plg.Args(
        "--resumption",
        help=(
            "scan for resumption support (SSL30 - TLS1.2) and for PSK support (TLS1.3)"
        ),
        action=utils.BooleanOptionalAction,
    )
    workers = [resumption.ScanResumption]


class GroupFeatures(plg.Plugin):
    """Argument group for features.
    """

    group = plg.Args(title="Feature to include into the scan")
    plugins = [
        ArgFeatures,
        ArgCompression,
        ArgDhGroup,
        ArgEncThenMac,
        ArgEphemKeyReuse,
        ArgExtMasterSecret,
        ArgFallback,
        ArgGrease,
        ArgHeartbeat,
        ArgOcspStapling,
        ArgRenegotiation,
        ArgResumption,
    ]

    @classmethod
    def args_parsed(cls, args, parser, subcommand, config):
        super().args_parsed(args, parser, subcommand, config)

        plugins = cls.plugins[1:]
        default = config.get(ArgFeatures.config.name)
        if default is None:
            default = not any(config.get(feat.config.name) for feat in plugins)

        for feat in plugins:
            val = config.get(feat.config.name)
            if val is None:
                config.set(feat.config.name, default)
                if default and feat.workers:
                    for worker in feat.workers:
                        plg.WorkManager.register(worker)


class ArgVulnerabilities(plg.Plugin):
    """Argument for Vulnerabilities.
    """

    config = structs.ConfigItem("vulnerabilities", type=bool, default=True)
    cli_args = plg.Args(
        "--vulnerabilities",
        help=(
            "specifies whether to include or exclude all vulnerabilities in the "
            "scan. Per vulnerability this behavior can be overruled by its "
            "specific command line option below. Defaults to true if no specific "
            "vulnerability is enabled, otherwise it defaults to false."
        ),
        action=utils.BooleanOptionalAction,
    )


class ArgCcsInjection(plg.Plugin):
    """Argument for CCS-injection vulnerability.
    """

    config = structs.ConfigItem("ccs_injection", type=bool, default=None)
    cli_args = plg.Args(
        "--ccs-injection",
        help="scan for vulnerability CCS-injection (only TL1.0 - TLS1.2)",
        action=utils.BooleanOptionalAction,
    )
    workers = [ccs_injection.ScanCcsInjection]


class ArgHeartbleed(plg.Plugin):
    """Argument for heartbleed vulnerability.
    """

    config = structs.ConfigItem("heartbleed", type=bool, default=None)
    cli_args = plg.Args(
        "--heartbleed",
        help="scan for the Heartbleed vulnerability CVE-2014-0160",
        action=utils.BooleanOptionalAction,
    )
    workers = [heartbleed.ScanHeartbleed]


class ArgPaddingOracle(plg.Plugin):
    """Argument for CBC padding oracles.
    """

    config = structs.ConfigItem("padding_oracle", type=bool, default=None)
    cli_args = plg.Args(
        "--padding-oracle",
        help="scan for CBC padding oracles",
        action=utils.BooleanOptionalAction,
    )
    workers = [padding_oracle.ScanPaddingOracle]


class ArgPaddingOracleAccuracy(plg.Plugin):
    """Argument for oracle accuracy.
    """

    config = structs.ConfigItem("oracle_accuracy", type=str, default="medium")
    cli_args = plg.Args(
        "--oracle-accuracy",
        help=(
            "the accuracy of the scan for CBC padding oracles. "
            "low: scan application data records for each TLS version with minimal "
            "set of cipher suites (fastest). "
            "medium: scan application data records for each TLS version and cipher "
            "suite combination (slower). "
            "high: scan application data, handshake and alert records for each TLS "
            "version and cipher suite combination (slowest). "
            "Default is medium."
        ),
        choices=["low", "medium", "high"],
        default="medium",
    )


class ArgRobot(plg.Plugin):
    """Argument for ROBOT vulnerability.
    """

    config = structs.ConfigItem("robot", type=bool, default=None)
    cli_args = plg.Args(
        "--robot",
        help=(
            "scan for ROBOT vulnerability CVE-2017-13099, etc. (only TL1.0 - TLS1.2)"
        ),
        action=utils.BooleanOptionalAction,
    )
    workers = [robot.ScanRobot]


class GroupVulnerabilities(plg.Plugin):
    """Argument group vulnerabilities.
    """

    group = plg.Args(title="Vulnerabilities to scan for")
    plugins = [
        ArgVulnerabilities,
        ArgCcsInjection,
        ArgHeartbleed,
        ArgPaddingOracle,
        ArgPaddingOracleAccuracy,
        ArgRobot,
    ]

    @classmethod
    def args_parsed(cls, args, parser, subcommand, config):
        super().args_parsed(args, parser, subcommand, config)

        plugins = cls.plugins[:]
        plugins.remove(ArgVulnerabilities)
        plugins.remove(ArgPaddingOracleAccuracy)

        default = config.get(ArgVulnerabilities.config.name)
        if default is None:
            default = not any(config.get(vuln.config.name) for vuln in plugins)

        base_vuln = False
        for vuln in plugins:
            val = config.get(vuln.config.name)
            if val is None:
                config.set(vuln.config.name, default)
                if default and vuln.workers:
                    for worker in vuln.workers:
                        plg.WorkManager.register(worker)
                        base_vuln = True

        if base_vuln:
            plg.WorkManager.register(base_vulnerabilities.ScanBaseVulnerabilities)


@plg.BaseCommand.extend
class SubcommandScan(plg.Plugin):
    """CLI plugin to perform a scan against a TLS server.
    """

    subcommand = plg.Args("scan", help="performs a TLS server scan")
    plugins = [
        GroupBasicScan,
        GroupX509,
        GroupTlsVersions,
        GroupFeatures,
        GroupVulnerabilities,
        GroupServerProfile,
    ]
    workers = [
        scanner_info.ScanStart,
        eval_cipher_suites.ScanCipherSuites,
        supported_groups.ScanSupportedGroups,
        sig_algo.ScanSigAlgs,
        scanner_info.ScanEnd,
    ]



# -*- coding: utf-8 -*-
"""Module containing the class for the client state
"""
# import basic stuff
from dataclasses import dataclass, field
from typing import List, Optional, Union, Mapping, Callable, Dict

# import own stuff
import tlsmate.key_exchange as kex
import tlsmate.recorder as rec
import tlsmate.structs as structs
import tlsmate.tls as tls


@dataclass
class ClientProfile(object):
    """Data class defining the TLS profile of the client.

    The client profile defines the protocol parameters, extensions and features
    which are supported by the client. The data is used when ``tlsmate`` generates
    the ClientHello message autonomously, e.g., when using::

        conn.send(msg.ClientHello)

        or

        conn.handshake()

    The client profile is used as well to check a received ServerHello for
    consistency, e.g., the version provided by the server must be in the
    version list of the client profile.

    Attributes:
        versions: The list of protocol versions the client shall support. Note,
            that the highest version will be picked to offer it in a
            ClientHello. For the extension SupportedVersions this list will be
            ordered, so that the highest version comes first. Arbitrary integer
            values can be provided in the list as well, allowing to check if
            the server ignores unknown values.
            Default: []
        compression_methods: A list of supported compression methods. This list
            will be used to populate the compression list in the ClientHello
            message.
            Default: [:obj:`tlsmate.tls.CompressionMethod.NULL`]
        cipher_suites: A list of cipher suites which will be offered to the
            server in the sequence given. Note, that arbitrary integer values
            are supported as well, allowing to check if the server ignores
            unknown values.
            Default: []
        support_sni: An indication if the SNI extension shall be supported. If
            True, the SNI will be taken from the CLI parameter ``--sni`` or (if
            not given), from the host name.
            Default: True
        ec_point_formats: The list of ec-point formats supported by the client.
            If set to None, the extension will not be generated.
            Default: None
        supported_groups: The list of named groups supported by the client. If
            set to None, the extension will not be present in the ClientHello
            message. Note, that arbitrary integer values are supported as well,
            allowing to check if the server ignores unknown values.
            Default: None
        signature_algorithms: The list of signature algorithms supported by the
            client. If set to None, the extension will not be present in the
            ClientHello message. Note, that arbitrary integer values are
            supported as well, allowing to check if the server ignores unknown
            values.
            Default: None
        heartbeat_mode: The mode which is offered in the heartbeat extension.
            If set to None, the extension will not be setup when using the
            :meth:`Client.client_hello` method.
            Default: None
        support_session_id: An indication if the client shall support
            resumption via the session id. Received session ids from the server
            will be offered in subsequent handshakes. Default: False
        support_session_ticket: An indication if the client shall support
            resumption via the extensions SessionTicket. Received session
            tickets from the server will be offered in subsequent handshakes.
            Default: False
        support_extended_master_secret: An indication if the client shall
            support the extensions ExtendedMasterSecret.
            Default: False
        support_encrypt_then_mac: An indication if the client shall support the
            EncryptThenMac extension.
            Default: False
        support_secure_renegotiation: An indication if the client shall support
            secure renegotiation. This will generate the RenegotiationInfo
            extension.
            Default: False
        support_scsv_renegotiation: An indication, if the cipher suite value
            TLS_EMPTY_RENEGOTIATION_INFO_SCSV shall be added to the cipher
            suite list. Only applicable if support_secure_renegotiation is
            True.
            Default: False
        support_psk: An indication whether the client offers a PSK with the
            ClientHello (i.e. NewSessionTicket message have been received
            before).
            Default: False
        support_status_request: An indication, if the extensions status request
            shall be supported.
            Default: False
        support_status_request_v2: The status type of the request. NONE is used
            to suppress the extension.
            Default: NONE
        key_shares: The list of key shares supported for TLS1.3. Note, that
            arbitrary integer values are supported as well, allowing to check
            if the server ignores unknown values.
            Default: None
        psk_key_exchange_modes: The list of PSK key exchange modes used in the
            extension psk_key_exchange_modes. Note, that arbitrary integer
            values are supported as well, allowing to check if the server
            ignores unknown values.
            Default: None
        early_data: The application data to be sent with 0-RTT. TLS1.3 only. If
            None, then no early data will be sent. Early data can only be sent
            in subsequent abbreviated handshakes.
            Default: None
        client_auth_supported: True, if client key and client certificate is present.
    """

    # common for all versions
    versions: List = field(default_factory=lambda: [])
    compression_methods: List = field(
        default_factory=lambda: [tls.CompressionMethod.NULL]
    )
    cipher_suites: List = field(default_factory=lambda: [])

    support_sni: bool = True
    ec_point_formats: Optional[List[Union[tls.EcPointFormat, int]]] = None
    supported_groups: Optional[List[tls.SupportedGroups]] = None
    signature_algorithms: Optional[List[Union[tls.SignatureScheme, int]]] = None
    heartbeat_mode: Optional[tls.HeartbeatMode] = None
    support_status_request: bool = False
    support_status_request_v2: tls.StatusType = tls.StatusType.NONE
    client_auth_supported: bool = False

    # TLS1.2 and below
    support_session_id: bool = False
    support_session_ticket: bool = False
    support_extended_master_secret: bool = False
    support_encrypt_then_mac: bool = False
    support_secure_renegotiation: bool = False
    support_scsv_renegotiation: bool = False

    # TLS1.3 specific
    support_psk: bool = False
    key_shares: Optional[List[tls.SupportedGroups]] = None
    psk_key_exchange_modes: Optional[List[tls.PskKeyExchangeMode]] = None
    early_data: Optional[bytes] = None


def _client_profile_interoperability() -> ClientProfile:
    """Define profile for interoperability, like used in modern browsers

    Profile properties:
      - TLS Versions 1.0 - 1.3
      - ECDHE cipher & RSA-based key transport
      - AESGCM, AES, CHACHA_POLY and 3DES as last resort
      - Signature algorithms: ECDSA+SHA1, RSA PKCS1+SHA1 as last resort
      - Resumption, encrypt-then-mac, extended-master-secret
      - pskmode psk_dhe
    """
    return ClientProfile(
        versions=[
            tls.Version.TLS10,
            tls.Version.TLS11,
            tls.Version.TLS12,
            tls.Version.TLS13,
        ],
        cipher_suites=[
            tls.CipherSuite.TLS_AES_128_GCM_SHA256,
            tls.CipherSuite.TLS_CHACHA20_POLY1305_SHA256,
            tls.CipherSuite.TLS_AES_256_GCM_SHA384,
            tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,
            tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256,
            tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256,
            tls.CipherSuite.TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256,
            tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,
            tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384,
            tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA,
            tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA,
            tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA,
            tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA,
            tls.CipherSuite.TLS_RSA_WITH_AES_128_GCM_SHA256,
            tls.CipherSuite.TLS_RSA_WITH_AES_256_GCM_SHA384,
            tls.CipherSuite.TLS_RSA_WITH_AES_128_CBC_SHA,
            tls.CipherSuite.TLS_RSA_WITH_AES_256_CBC_SHA,
            tls.CipherSuite.TLS_RSA_WITH_3DES_EDE_CBC_SHA,
        ],
        supported_groups=[
            tls.SupportedGroups.X25519,
            tls.SupportedGroups.SECP256R1,
            tls.SupportedGroups.SECP384R1,
            tls.SupportedGroups.SECP521R1,
        ],
        signature_algorithms=[
            tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
            tls.SignatureScheme.RSA_PKCS1_SHA256,
            tls.SignatureScheme.ECDSA_SECP384R1_SHA384,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
            tls.SignatureScheme.RSA_PKCS1_SHA384,
            tls.SignatureScheme.ECDSA_SECP521R1_SHA512,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
            tls.SignatureScheme.RSA_PKCS1_SHA512,
            tls.SignatureScheme.ECDSA_SHA1,
            tls.SignatureScheme.RSA_PKCS1_SHA1,
        ],
        key_shares=[tls.SupportedGroups.X25519, tls.SupportedGroups.SECP256R1],
        psk_key_exchange_modes=[tls.PskKeyExchangeMode.PSK_DHE_KE],
        ec_point_formats=[tls.EcPointFormat.UNCOMPRESSED],
        support_extended_master_secret=True,
        support_session_id=True,
        support_session_ticket=True,
        support_encrypt_then_mac=True,
    )


def _client_profile_legacy():
    """Define profile for legacy like client

    Profile properties:
      - TLS Versions 1.0 - 1.2
      - ECDHE cipher & DHE & RSA-based key transport
      - AESGCM, AES, CHACHA_POLY and 3DES as last resort
      - Signature algorithms: ECDSA+SHA1, RSA PKCS1+SHA1 as last resort
      - Resumption, encrypt-then-mac, extended-master-secret
    """

    return ClientProfile(
        versions=[tls.Version.TLS10, tls.Version.TLS11, tls.Version.TLS12],
        cipher_suites=[
            tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,
            tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256,
            tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256,
            tls.CipherSuite.TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256,
            tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,
            tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384,
            tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA,
            tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA,
            tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA,
            tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA,
            tls.CipherSuite.TLS_DHE_RSA_WITH_AES_128_GCM_SHA256,
            tls.CipherSuite.TLS_DHE_RSA_WITH_AES_256_GCM_SHA384,
            tls.CipherSuite.TLS_DHE_RSA_WITH_CHACHA20_POLY1305_SHA256,
            tls.CipherSuite.TLS_DHE_RSA_WITH_AES_128_CBC_SHA256,
            tls.CipherSuite.TLS_DHE_RSA_WITH_AES_256_CBC_SHA256,
            tls.CipherSuite.TLS_DHE_RSA_WITH_AES_128_CBC_SHA,
            tls.CipherSuite.TLS_DHE_RSA_WITH_AES_256_CBC_SHA,
            tls.CipherSuite.TLS_RSA_WITH_AES_128_GCM_SHA256,
            tls.CipherSuite.TLS_RSA_WITH_AES_256_GCM_SHA384,
            tls.CipherSuite.TLS_RSA_WITH_AES_128_CBC_SHA,
            tls.CipherSuite.TLS_RSA_WITH_AES_256_CBC_SHA,
            tls.CipherSuite.TLS_RSA_WITH_3DES_EDE_CBC_SHA,
        ],
        supported_groups=[
            tls.SupportedGroups.X25519,
            tls.SupportedGroups.SECP256R1,
            tls.SupportedGroups.SECP384R1,
            tls.SupportedGroups.SECP521R1,
        ],
        signature_algorithms=[
            tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
            tls.SignatureScheme.RSA_PKCS1_SHA256,
            tls.SignatureScheme.ECDSA_SECP384R1_SHA384,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
            tls.SignatureScheme.RSA_PKCS1_SHA384,
            tls.SignatureScheme.ECDSA_SECP521R1_SHA512,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
            tls.SignatureScheme.RSA_PKCS1_SHA512,
            tls.SignatureScheme.ECDSA_SHA1,
            tls.SignatureScheme.RSA_PKCS1_SHA1,
        ],
        ec_point_formats=[tls.EcPointFormat.UNCOMPRESSED],
        support_extended_master_secret=True,
        support_session_id=True,
        support_session_ticket=True,
        support_encrypt_then_mac=True,
    )


def _client_profile_modern():
    """Define profile for "modern" configurations

    Profile properties:
      - TLS Versions 1.2 + 1.3
      - ECDHE cipher
      - AESGCM, CHACHA_POLY
      - signatures: ECDSA + PSS_RSAE
      - Resumption, encrypt-then-mac, extended-master-secret
      - pskmode psk_dhe
    """

    return ClientProfile(
        versions=[tls.Version.TLS12, tls.Version.TLS13],
        cipher_suites=[
            tls.CipherSuite.TLS_AES_128_GCM_SHA256,
            tls.CipherSuite.TLS_CHACHA20_POLY1305_SHA256,
            tls.CipherSuite.TLS_AES_256_GCM_SHA384,
            tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,
            tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256,
            tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256,
            tls.CipherSuite.TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256,
            tls.CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,
            tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384,
        ],
        supported_groups=[
            tls.SupportedGroups.X25519,
            tls.SupportedGroups.SECP256R1,
            tls.SupportedGroups.SECP384R1,
            tls.SupportedGroups.SECP521R1,
        ],
        signature_algorithms=[
            tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
            tls.SignatureScheme.ECDSA_SECP384R1_SHA384,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
            tls.SignatureScheme.ECDSA_SECP521R1_SHA512,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
        ],
        key_shares=[tls.SupportedGroups.X25519, tls.SupportedGroups.SECP256R1],
        psk_key_exchange_modes=[tls.PskKeyExchangeMode.PSK_DHE_KE],
        ec_point_formats=[tls.EcPointFormat.UNCOMPRESSED],
        support_sni=True,
        support_extended_master_secret=True,
        support_session_id=True,
        support_session_ticket=True,
    )


def _client_profile_tls13():
    """Define profile for TLS1.3 only.

    Profile properties:
      - TLS Version 1.3
      - AESGCM + CHACHA_POLY
      - pskmode psk_dhe
    """

    return ClientProfile(
        versions=[tls.Version.TLS13],
        cipher_suites=[
            tls.CipherSuite.TLS_AES_128_GCM_SHA256,
            tls.CipherSuite.TLS_CHACHA20_POLY1305_SHA256,
            tls.CipherSuite.TLS_AES_256_GCM_SHA384,
        ],
        supported_groups=[
            tls.SupportedGroups.X25519,
            tls.SupportedGroups.SECP256R1,
            tls.SupportedGroups.SECP384R1,
            tls.SupportedGroups.SECP521R1,
        ],
        signature_algorithms=[
            tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
            tls.SignatureScheme.RSA_PKCS1_SHA256,
            tls.SignatureScheme.ECDSA_SECP384R1_SHA384,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
            tls.SignatureScheme.RSA_PKCS1_SHA384,
            tls.SignatureScheme.ECDSA_SECP521R1_SHA512,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
            tls.SignatureScheme.RSA_PKCS1_SHA512,
        ],
        key_shares=[tls.SupportedGroups.X25519, tls.SupportedGroups.SECP256R1],
        psk_key_exchange_modes=[tls.PskKeyExchangeMode.PSK_DHE_KE],
        ec_point_formats=[tls.EcPointFormat.UNCOMPRESSED],
        support_sni=True,
    )


def create_profile(profile: tls.Profile) -> ClientProfile:
    """Creates a client profile according to the given profile.

    The following profiles are supported:

    - :obj:`tlsmate.tls.Profile.INTEROPERABILITY`

      - TLS Versions 1.0 - 1.3
      - ECDHE cipher & RSA-based key transport
      - AESGCM, AES, CHACHA_POLY and 3DES as last resort
      - Signature algorithms: ECDSA+SHA1, RSA PKCS1+SHA1 as last resort
      - Resumption, encrypt-then-mac, extended-master-secret
      - pskmode psk_dhe

    - :obj:`tlsmate.tls.Profile.LEGACY`

      - TLS Versions 1.0 - 1.2
      - ECDHE cipher & DHE & RSA-based key transport
      - AESGCM, AES, CHACHA_POLY and 3DES as last resort
      - Signature algorithms: ECDSA+SHA1, RSA PKCS1+SHA1 as last resort
      - Resumption, encrypt-then-mac, extended-master-secret

    - :obj:`tlsmate.tls.Profile.MODERN`

      - TLS Versions 1.2 + 1.3
      - ECDHE cipher
      - AESGCM, CHACHA_POLY
      - signatures: ECDSA + PSS_RSAE
      - Resumption, encrypt-then-mac, extended-master-secret
      - pskmode psk_dhe

    - :obj:`tlsmate.tls.Profile.TLS13`

      - TLS Version 1.3
      - AESGCM + CHACHA_POLY
      - pskmode psk_dhe

    Arguments:
        profile (:obj:`tlsmate.tls.Profile`): the profile to which the client
            shall be initialized.
    """

    mapping: Mapping[tls.Profile, Callable[[], ClientProfile]] = {
        tls.Profile.INTEROPERABILITY: _client_profile_interoperability,
        tls.Profile.TLS13: _client_profile_tls13,
        tls.Profile.MODERN: _client_profile_modern,
        tls.Profile.LEGACY: _client_profile_legacy,
    }

    if profile not in mapping:
        raise ValueError(f"client profile {profile} unknown")

    return mapping[profile]()


def init_profile(
    profile_values: Optional[structs.ProfileValues] = None,
) -> ClientProfile:
    """Creates a client profile reset to basic settings

    :note:
        At least the versions and the cipher_suite list must be provided before
        this profile can be used.

    Compression methods is set to [:obj:`tlsmate.tls.CompressionMethod.NULL`], and
    by default the sni extension is enabled. Everything else is empty or disabled.

    Arguments:
        profile_values: the profile values to additionally use to
            initialize the client profile
    """

    profile = ClientProfile()
    if profile_values:
        profile.versions = profile_values.versions[:]
        profile.cipher_suites = profile_values.cipher_suites[:]
        if profile_values.supported_groups:
            profile.supported_groups = profile_values.supported_groups[:]

        if profile_values.signature_algorithms:
            profile.signature_algorithms = profile_values.signature_algorithms[:]

        if profile_values.key_shares:
            profile.key_shares = profile_values.key_shares[:]

    return profile


class SessionState(object):
    """Represents the data associated with the targe.

    The lifetime of this object exceeds the lifetime of a connection.
    """

    def __init__(
        self, host: str, port: int, sni: Optional[str], recorder: rec.Recorder
    ) -> None:
        self.recorder = recorder
        self.host: str = host
        self.port: int = port
        self.sni: Optional[str] = sni
        self.session_state_id: Optional[structs.SessionStateId] = None
        self.session_state_ticket: Optional[structs.SessionStateTicket] = None
        self.psks: List[structs.Psk] = []
        self.key_shares: Dict[tls.SupportedGroups, kex.KeyExchange] = {}
        self.version: Optional[tls.Version] = None
        self.cs_details: Optional[structs.CipherSuiteDetails] = None
        self.server_issues: List[structs.Malfunction] = []

    def save_session_state_id(self, session_state: structs.SessionStateId) -> None:
        """Saves a session state

        Args:
            session_state: A session state to be stored on the client level,
                usable to resume connections using the session_id
        """

        self.session_state_id = session_state

    def get_session_state_id(self) -> Optional[structs.SessionStateId]:
        """Get the session state (id)

        Returns:
            the session state to resume a session from
        """

        return self.session_state_id

    def save_session_state_ticket(
        self, session_state: structs.SessionStateTicket
    ) -> None:
        """Saves a session state

        Args:
            session_state: A session state to be stored on the client level,
                usable to resume connections using the session ticket.
        """

        self.session_state_ticket = session_state

    def get_session_state_ticket(self) -> Optional[structs.SessionStateTicket]:
        """Get the session state (ticket)

        Returns:
            the session state to resume a session from
        """

        return self.session_state_ticket

    def save_psk(self, psk: structs.Psk) -> None:
        """Save a TLS1.3 PSK

        Arguments:
            psk: A pre-shared key be stored on the client level, usable to
                resume connections using the pre-shared key extension.
        """

        self.psks.append(psk)

    def report_server_issue(
        self,
        issue: tls.ServerIssue,
        message: Optional[tls.HandshakeType] = None,
        extension: Optional[tls.Extension] = None,
    ) -> None:
        """Store a server issue, if not done

        Arguments:
            issue: the reason for the exception
            message: the message, if applicable
            extension: the extension, if applicable
        """

        malfunction = structs.Malfunction(
            issue=issue, message=message, extension=extension
        )

        if malfunction not in self.server_issues:
            self.server_issues.append(malfunction)

    def create_key_share(self, group: tls.SupportedGroups) -> bytes:
        """Provide the key share for a given group.

        Arguments:
            group: the group to create a key share for

        Returns:
            the key to exchange with the remote side
        """

        key_share = kex.instantiate_named_group(self.recorder, group)
        self.key_shares[group] = key_share
        return key_share.get_key_share()



# -*- coding: utf-8 -*-
"""Module defining some utilities related to certificates
"""

# import basic stuff
import stringprep
import unicodedata
from typing import List

# import own stuff
import tlsmate.tls as tls

# import other stuff
from cryptography.x509.oid import SignatureAlgorithmOID as sigalg_oid
from cryptography.x509 import NameAttribute, RelativeDistinguishedName, Name


def string_prep(label: str) -> str:
    """Prepares a string for comparison according to RFC 3435.

    Copy&Pasted code. Let's hope the best.

    Arguments:
        labelThe string to prepare

    Returns:
        the prepared string
    """
    # Map
    newlabel = []
    for c in label:
        if stringprep.in_table_b1(c):
            # Map to nothing
            continue
        newlabel.append(stringprep.map_table_b2(c))
    label = "".join(newlabel)

    # Normalize
    label = unicodedata.normalize("NFKC", label)

    # Prohibit
    for c in label:
        if (
            stringprep.in_table_c12(c)
            or stringprep.in_table_c22(c)
            or stringprep.in_table_c3(c)
            or stringprep.in_table_c4(c)
            or stringprep.in_table_c5(c)
            or stringprep.in_table_c6(c)
            or stringprep.in_table_c7(c)
            or stringprep.in_table_c8(c)
            or stringprep.in_table_c9(c)
        ):
            raise UnicodeError("Invalid character %r" % c)

    # Check bidi
    RandAL = map(stringprep.in_table_d1, label)
    for c in RandAL:  # type: ignore
        if c:
            # There is a RandAL char in the string. Must perform further
            # tests:
            # 1) The characters in section 5.8 MUST be prohibited.
            # This is table C.8, which was already checked
            # 2) If a string contains any RandALCat character, the string
            # MUST NOT contain any LCat character.
            if filter(stringprep.in_table_d2, label):
                raise UnicodeError("Violation of BIDI requirement 2")

            # 3) If a string contains any RandALCat character, a
            # RandALCat character MUST be the first character of the
            # string, and a RandALCat character MUST be the last
            # character of the string.
            if not RandAL[0] or not RandAL[-1]:  # type: ignore
                raise UnicodeError("Violation of BIDI requirement 3")

    return label


def remove_subdomain(name: str) -> str:
    """Removes the subdomain from a given URL.

    Example:
        >>> remove_subdomain("www.example.com")
        '.example.com'

    Arguments:
        name: the domain

    Returns:
        the string with the subdomain removed. The string starts with the dot.
    """
    try:
        pos = name.index(".")
    except ValueError:
        return name
    return name[pos:]


def subject_matches(subject: str, full_domain: str, name_no_subdomain: str) -> bool:
    """Checks, if a given subject matches either a full domain or a wildcard domain.

    The subject is prepared for string comparison first.

    Arguments:
        subject: the subject to be checked
        full_domain: the full domain, e.g. "www.example.com". This argument
            should be string-prepped.
        name_no_subdomain: the domain without the leading subdomain, e.g.
            ".example.com". This argument should be string_prepped.

    Returns:
        True, if the subject matches either the full domain or the
        name_no_subdomain
    """
    subject = string_prep(subject)
    if subject.startswith("*"):
        return subject[1:] == name_no_subdomain
    return subject == full_domain


def equal_oid(
    name_attrs1: List[NameAttribute], name_attrs2: List[NameAttribute]
) -> bool:
    """Check two attributes for equality

    Arguments:
        name_attrs1: a list of attribute name objects as defined by the library
            crypthography.
        name_attrs2: a list of attribute name objects as defined by the library
            crypthography.

    Returns:
        True, if both attributes are equal as defined in RFC 5280.
    """

    values1 = [string_prep(name_attr.value) for name_attr in name_attrs1]
    values2 = [string_prep(name_attr.value) for name_attr in name_attrs2]
    return set(values1) == set(values2)


def equal_rdn(rdn1: RelativeDistinguishedName, rdn2: RelativeDistinguishedName) -> bool:
    """Check two rdns (Relative Distinguished Name) for equality

    Arguments:
        rdn1: an rdn object as defined by the library crypthography.
        rdn2: an rdn object as defined by the library crypthography.

    Returns:
        True, if both rdns are equal as defined in RFC 5280.
    """
    return all(
        equal_oid(
            rdn1.get_attributes_for_oid(name_attr.oid),
            rdn2.get_attributes_for_oid(name_attr.oid),
        )
        for name_attr in rdn1
    )


def equal_names(name1: Name, name2: Name) -> bool:
    """Check two x509 names for equality

    Arguments:
        name1 (:obj:`cryptography.x509.Name`): a name object as defined by the
            library crypthography.
        name2 (:obj:`cryptography.x509.Name`): an name object as defined by the
            library crypthography.

    Returns:
        bool: True, if both names are equal as defined in RFC 5280.
    """
    if len(name1.rdns) != len(name2.rdns):  # type: ignore
        return False

    return all(equal_rdn(rdn1, rdn2) for rdn1, rdn2 in zip(name1.rdns, name2.rdns))


_sig_alg = {
    sigalg_oid.RSA_WITH_MD5: tls.SignatureScheme.RSA_PKCS1_MD5,
    sigalg_oid.RSA_WITH_SHA1: tls.SignatureScheme.RSA_PKCS1_SHA1,
    sigalg_oid.RSA_WITH_SHA224: tls.SignatureScheme.RSA_PKCS1_SHA224,
    sigalg_oid.RSA_WITH_SHA256: tls.SignatureScheme.RSA_PKCS1_SHA256,
    sigalg_oid.RSA_WITH_SHA384: tls.SignatureScheme.RSA_PKCS1_SHA384,
    sigalg_oid.RSA_WITH_SHA512: tls.SignatureScheme.RSA_PKCS1_SHA512,
    sigalg_oid.ECDSA_WITH_SHA1: tls.SignatureScheme.ECDSA_SHA1,
    sigalg_oid.ECDSA_WITH_SHA224: tls.SignatureScheme.ECDSA_SECP224R1_SHA224,
    sigalg_oid.ECDSA_WITH_SHA256: tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
    sigalg_oid.ECDSA_WITH_SHA384: tls.SignatureScheme.ECDSA_SECP384R1_SHA384,
    sigalg_oid.ECDSA_WITH_SHA512: tls.SignatureScheme.ECDSA_SECP521R1_SHA512,
    sigalg_oid.DSA_WITH_SHA1: tls.SignatureScheme.DSA_SHA1,
    sigalg_oid.DSA_WITH_SHA224: tls.SignatureScheme.DSA_SHA224,
    sigalg_oid.DSA_WITH_SHA256: tls.SignatureScheme.DSA_SHA256,
    sigalg_oid.ED25519: tls.SignatureScheme.ED25519,
    sigalg_oid.ED448: tls.SignatureScheme.ED448,
}

_pss_sig_alg = {
    "sha256": tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
    "sha384": tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
    "sha512": tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
}


def map_x509_sig_scheme(x509_hash, x509_oid):
    """Maps a given x509 hash and oid to the internal signature scheme.

    Arguments:
        x509_hash (:obj:`cryptography.hazmat.primitives.hashes.Hash`): the hash object
        x509_oid (:obj:`cryptography.x509.ObjectIdentifier`): the oid for the signature
            algorithm

    Returns:
        :class:`tlsmate.tls.SignatureScheme`: the signature scheme
    """
    if x509_oid is sigalg_oid.RSASSA_PSS:
        sig_scheme = _pss_sig_alg(x509_hash.name)
        if sig_scheme is None:
            raise ValueError(f"signature algorithm {x509_hash} not supported")
        return sig_scheme
    else:
        sig_scheme = _sig_alg.get(x509_oid)
        if sig_scheme is None:
            raise ValueError(f"signature algorithm {x509_oid} not supported")
        return sig_scheme



# -*- coding: utf-8 -*-
"""Module scanning for the ROBOT vulnerability

Refer to CVE-2017-13099, etc.

Padding oracle for RSA-based key transport, refer to https://robotattack.org
"""
# import basic stuff
import math

# import own stuff
import tlsmate.msg as msg
import tlsmate.plugin as plg
import tlsmate.tls as tls
import tlsmate.utils as utils

# import other stuff


def _rsa_encrypt(msg, e, n, mod_bytes):
    return int(pow(msg, e, n)).to_bytes(mod_bytes, byteorder="big")


class ScanRobot(plg.Worker):
    name = "robot"
    descr = "scan for ROBOT vulnerability"
    prio = 41

    def _get_oracle_results(self, with_ccs):
        def cke_pre_serialization(message):
            message.rsa_encrypted_pms = self.enc_pms

        results = []
        for self.enc_pms in self._rsa_encrypted_pms:
            with self.client.create_connection() as conn:
                conn.send(msg.ClientHello)
                conn.wait(msg.ServerHello)
                conn.wait(msg.Certificate)
                conn.wait(msg.CertificateRequest, optional=True)
                conn.wait(msg.ServerHelloDone)
                conn.send(
                    msg.ClientKeyExchange, pre_serialization=cke_pre_serialization
                )
                self.premaster_secret = self.rnd_pms
                if with_ccs:
                    conn.send(msg.ChangeCipherSpec)
                    conn.send(msg.Finished)

                try:
                    rec_msg, rec_bytes = conn.wait_msg_bytes(msg.Any, timeout=1000)
                    results.append(hash(bytes(rec_bytes)))

                except Exception as exc:
                    results.append(hash(str(exc)))

        return results

    def _determine_status(self):
        for send_ccs_finished in [True, False]:
            results = self._get_oracle_results(send_ccs_finished)
            if len(set(results)) == 1:
                continue

            results2 = self._get_oracle_results(send_ccs_finished)
            for res1, res2 in zip(results, results2):
                if res1 != res2:
                    return tls.RobotVulnerability.INCONSITENT_RESULTS

            if results[1] == results[2] == results[3]:
                return tls.RobotVulnerability.WEAK_ORACLE
            return tls.RobotVulnerability.STRONG_ORACLE
        return tls.RobotVulnerability.NOT_VULNERABLE

    def run(self):
        values = self.server_profile.get_profile_values(
            [tls.Version.TLS10, tls.Version.TLS11, tls.Version.TLS12], full_hs=True
        )
        rsa_ciphers = utils.filter_cipher_suites(
            values.cipher_suites, key_algo=[tls.KeyExchangeAlgorithm.RSA]
        )
        if rsa_ciphers:
            self.client.init_profile(profile_values=values)
            self.client.profile.cipher_suites = rsa_ciphers
            with self.client.create_connection() as conn:
                conn.handshake()

            if not conn.handshake_completed:
                status = tls.RobotVulnerability.UNDETERMINED

            else:
                cert = conn.msg.server_certificate.chain.certificates[0]
                pub_nbrs = cert.parsed.public_key().public_numbers()
                modulus_bits = int(math.ceil(math.log(pub_nbrs.n, 2)))
                modulus_bytes = (modulus_bits + 7) // 8
                pad_len = (modulus_bytes - 48 - 3) * 2
                rnd_pad = ("abcd" * (pad_len // 2 + 1))[:pad_len]
                self.rnd_pms = (
                    "aa11223344556677889911223344556677889911223344"
                    "5566778899112233445566778899112233445566778899"
                )
                pms_good_in = int("0002" + rnd_pad + "00" + "0303" + self.rnd_pms, 16)
                # wrong first two bytes
                pms_bad_in1 = int("4117" + rnd_pad + "00" + "0303" + self.rnd_pms, 16)
                # 0x00 on a wrong position, also trigger older JSSE bug
                pms_bad_in2 = int("0002" + rnd_pad + "11" + self.rnd_pms + "0011", 16)
                # no 0x00 in the middle
                pms_bad_in3 = int("0002" + rnd_pad + "11" + "1111" + self.rnd_pms, 16)
                # wrong version number (according to Klima / Pokorny / Rosa paper)
                pms_bad_in4 = int("0002" + rnd_pad + "00" + "0202" + self.rnd_pms, 16)

                self._rsa_encrypted_pms = [
                    _rsa_encrypt(pms, pub_nbrs.e, pub_nbrs.n, modulus_bytes)
                    for pms in [
                        pms_good_in,
                        pms_bad_in1,
                        pms_bad_in2,
                        pms_bad_in3,
                        pms_bad_in4,
                    ]
                ]
                status = self._determine_status()

        else:
            status = tls.RobotVulnerability.NOT_APPLICABLE

        self.server_profile.allocate_vulnerabilities()
        self.server_profile.vulnerabilities.robot = status



# -*- coding: utf-8 -*-
"""Module containing a classes to abstract the socket
"""
# import basic stuff
import socket
import select
import logging
import time
import sys
import socks  # type: ignore
import urllib.parse
from typing import Tuple, Optional

# import own stuff
import tlsmate.config as conf
import tlsmate.recorder as rec
import tlsmate.resolver as resolver
import tlsmate.structs as structs
import tlsmate.tls as tls
import tlsmate.utils as utils

# import other stuff


class Socket(object):
    """Class implementing the socket interface.

    Arguments:
        tlsmate: the application object
    """

    def __init__(self, config: conf.Configuration, recorder: rec.Recorder) -> None:
        self._socket: Optional[socks.socksocket] = None
        self._config = config
        self._recorder = recorder
        self._fragment_max_size = 16384

    def open_socket(self, l4_addr: structs.TransportEndpoint) -> None:
        """Opens a socket.

        Arguments:
            l4_addr: The L4-endpoint, consisting of the IP-address and the
                port.
        """

        addr_info: Tuple
        if l4_addr.host_type is tls.HostType.HOST:
            l4_addr = resolver.get_ip_endpoint(
                l4_addr,
                proxy=self._config.get("proxy"),
                ipv6_preference=self._config.get("ipv6_preference"),
            )

        addr_info = (l4_addr.host, l4_addr.port)
        family = (
            socket.AF_INET
            if l4_addr.host_type is tls.HostType.IPV4
            else socket.AF_INET6
        )

        if self._recorder.is_injecting():
            return

        try:
            self._socket = socks.socksocket(family, socket.SOCK_STREAM)
            proxy = self._config.get("proxy")
            if proxy:
                parsed = urllib.parse.urlparse(proxy)
                self._socket.set_proxy(socks.HTTP, parsed.hostname, parsed.port)
            self._socket.settimeout(5.0)
            self._socket.connect(addr_info)
            self._socket.settimeout(None)
            addr = self._socket.getsockname()
            laddr = addr[0]
            lport = addr[1]
            addr = self._socket.getpeername()
            raddr = addr[0]
            rport = addr[1]

        except OSError:
            utils.exit_with_error(f"Cannot open TCP connection to {l4_addr}")

        if self._config.get("progress"):
            sys.stderr.write(".")
            sys.stderr.flush()

        logging.info(f"{utils.Log.time()}: Socket opened")
        logging.info(f"local address: {laddr}:{lport}")
        logging.info(f"remote address: {raddr}:{rport}")

    def close_socket(self) -> None:
        """Closes a socket.
        """

        if self._socket is not None:
            logging.info(f"{utils.Log.time()}: closing socket")
            self._socket.close()
            self._socket = None

    def sendall(self, data: bytes) -> None:
        """Sends data to the network.

        Arguments:
            data: The data to send.
        """

        cont = self._recorder.trace_socket_sendall(data)
        if cont:
            if self._socket is None:
                return

            self._socket.sendall(data)

    def recv_data(self, timeout: float = 5) -> bytes:
        """Wait for data from the network.

        Arguments:
            timeout: The maximum time to wait in seconds.

        Returns:
            The bytes received or None if the timeout expired.

        Raises:
            TlsMsgTimeoutError: If no data is received within the given timeout
            TlsConnectionClosedError: If the connection was closed.
        """

        data = self._recorder.inject_socket_recv()
        if data is None:
            assert self._socket

            start = time.time()
            rfds, wfds, efds = select.select([self._socket], [], [], timeout)
            timeout = time.time() - start
            if not rfds:
                self._recorder.trace_socket_recv(timeout, rec.SocketEvent.TIMEOUT)
                raise tls.TlsMsgTimeoutError

            try:
                data = self._socket.recv(self._fragment_max_size)

            except ConnectionResetError as exc:
                self._recorder.trace_socket_recv(timeout, rec.SocketEvent.CLOSURE)
                self.close_socket()
                raise tls.TlsConnectionClosedError(exc)

            self._recorder.trace_socket_recv(timeout, rec.SocketEvent.DATA, data=data)
        if data == b"":
            self.close_socket()
            raise tls.TlsConnectionClosedError()

        return data



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
from tlsmate.workers.eval_cipher_suites import ScanCipherSuites
from tlsmate.tlssuite import TlsSuiteTester
from tlsmate.tlssuite import TlsLibrary

ssl2_ck = [
    "SSL_CK_RC4_128_WITH_MD5",
    "SSL_CK_RC2_128_CBC_WITH_MD5",
    "SSL_CK_IDEA_128_CBC_WITH_MD5",
    "SSL_CK_DES_192_EDE3_CBC_WITH_MD5",
]


class TestCase(TlsSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the TcRecorder class.
    """

    sp_out_yaml = "profile_basic_ssl2"
    recorder_yaml = "recorder_eval_cipher_suites_ssl2"
    path = pathlib.Path(__file__)
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-rsa --cert2 server-ecdsa --no-cert-chain "
        "-- -www -cipher ALL -ssl2"
    )
    library = TlsLibrary.openssl1_0_2

    server = "localhost"

    def check_versions(self, versions):
        assert len(versions) == 6

        assert versions[0]["version"]["name"] == "SSL20"
        assert versions[0]["support"] == "TRUE"

        assert versions[1]["version"]["name"] == "SSL30"
        assert versions[1]["support"] == "FALSE"

        assert versions[2]["version"]["name"] == "TLS10"
        assert versions[2]["support"] == "FALSE"

        assert versions[3]["version"]["name"] == "TLS11"
        assert versions[3]["support"] == "FALSE"

        assert versions[4]["version"]["name"] == "TLS12"
        assert versions[4]["support"] == "FALSE"

        assert versions[5]["version"]["name"] == "TLS13"
        assert versions[5]["support"] == "FALSE"

        for a, b in zip(ssl2_ck, versions[0]["cipher_kinds"]):
            assert a == b["name"]

    def check_profile(self, profile):
        self.check_versions(profile["versions"])

    def run(self, tlsmate, is_replaying):
        for vers in ["sslv2", "sslv3", "tls10", "tls11", "tls12", "tls13"]:
            tlsmate.config.set(vers, True)
        server_profile = tlsmate.server_profile
        ScanCipherSuites(tlsmate).run()

        self.check_profile(server_profile.make_serializable())


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Module defining some utilities
"""
# import basic stuff
import time
import logging
import json
import os
import sys
import argparse
from typing import List, Optional, Type, Any, Tuple, Union

# import own stuff
import tlsmate.mappings as mappings
import tlsmate.pdu as pdu
import tlsmate.structs as structs
import tlsmate.tls as tls

# import other stuff
import yaml
import pathlib


class BooleanOptionalAction(argparse.Action):
    """Class to support --flag and --no-flag arguments.

    Will be natively supported by Python3.9
    """

    def __init__(
        self,
        option_strings: List[str],
        dest: str,
        default: Optional[str] = None,
        type: Type = None,
        choices: List[str] = None,
        required: bool = False,
        help: Optional[str] = None,
        metavar: Optional[str] = None,
    ) -> None:

        _option_strings = []
        for option_string in option_strings:
            _option_strings.append(option_string)

            if option_string.startswith("--"):
                option_string = "--no-" + option_string[2:]
                _option_strings.append(option_string)

        if help is not None and default is not None:
            help += f" (default: {default})"

        super().__init__(
            option_strings=_option_strings,
            dest=dest,
            nargs=0,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        if option_string in self.option_strings:
            setattr(namespace, self.dest, not option_string.startswith("--no-"))

    def format_usage(self) -> str:
        return " | ".join(self.option_strings)


def serialize_data(
    data: Any,
    file_name: Optional[str] = None,
    replace: bool = True,
    use_json: bool = False,
    indent: int = 4,
) -> None:
    """Serialize the data object to JSON or yaml.

    Arguments:
        data: the serializable data structure
        file_name: the file to write the serialized data to. If not given, the
            serialized object is printed on STDOUT.
        replace: If True, allow overwriting an existing file. Defaults to True.
            This argument is only evaluated if a file_name is given.
        json: If True, use JSON, else use Yaml.
        indent: The indentation to apply.
    """

    if file_name is not None:
        if not replace and pathlib.Path(file_name).exists():
            form = "JSON" if use_json else "Yaml"
            print(f"File {file_name} existing. {form}-file not generated")
            return

        with open(file_name, "w") as fd:
            if use_json:
                json.dump(data, fd, indent=indent, sort_keys=True)

            else:
                yaml.dump(data, fd, indent=indent)
    else:
        if use_json:
            print(json.dumps(data, indent=indent, sort_keys=True))

        else:
            print(yaml.dump(data, indent=indent))


def deserialize_data(file_name: str) -> Any:
    """Deserialize from a JSON- or Yaml-file.

    Arguments:
        file_name: the full file name

    Returns:
        the deserialized object
    """

    with open(file_name) as fd:
        return yaml.safe_load(fd)


def fold_string(text: str, max_length: int, sep: str = " ") -> List[str]:
    """Splits a string into lines of the given length.

    Arguments:
        text: the string to split
        length: the maximum length of the line
        sep: the separator where splitting the text is allowed

    Returns:
        list: the list of strings
    """

    ret_lines = []
    tokens: List[str] = []
    length = 0
    sep_len = len(sep)

    for token in text.split(sep):
        token_len = len(token)
        if length + token_len + sep_len > max_length:
            if tokens:
                ret_lines.append(sep.join(tokens) + sep)
                tokens = [token]
                length = token_len

            else:
                ret_lines.append(token + sep)
                tokens = []
                length = 0

        else:
            tokens.append(token)
            length += token_len + sep_len

    if tokens:
        ret_lines.append(sep.join(tokens))

    return ret_lines


def get_cipher_suite_details(
    cipher_suite: tls.CipherSuite,
) -> Optional[structs.CipherSuiteDetails]:
    """Get details for a given cipher suite

    Arguments:
        cipher_suite: The given cipher suite.

    Returns:
        The structure with the detailed info regarding the given cipher suite.
    """

    cs = mappings.supported_cipher_suites.get(cipher_suite)
    if cs is None:
        return None

    ciph = mappings.supported_ciphers[cs.cipher]
    key = mappings.key_exchange.get(cs.key_ex)
    if key is not None:
        key_ex_supp = key.key_ex_supported

    else:
        key_ex_supp = False

    mac_struct = mappings.supported_macs.get(cs.mac)

    return structs.CipherSuiteDetails(
        cipher_suite=cipher_suite,
        full_hs=(key_ex_supp and ciph.cipher_supported),
        key_exchange_supported=key_ex_supp,
        key_algo=cs.key_ex,
        key_algo_struct=key,
        cipher=cs.cipher,
        cipher_struct=ciph,
        mac=cs.mac,
        mac_struct=mac_struct,
    )


def filter_cipher_suites(
    cs_list: List[tls.CipherSuite],
    key_exch: Optional[List[tls.KeyExchangeType]] = None,
    key_auth: Optional[List[tls.KeyAuthentication]] = None,
    key_algo: Optional[List[tls.KeyExchangeAlgorithm]] = None,
    cipher_type: Optional[List[tls.CipherType]] = None,
    cipher_prim: Optional[List[tls.CipherPrimitive]] = None,
    cipher: Optional[List[tls.SymmetricCipher]] = None,
    mac: Optional[List[tls.HashPrimitive]] = None,
    version: Optional[tls.Version] = None,
    full_hs: Optional[bool] = None,
    key_exchange_supported: Optional[bool] = None,
    remove: bool = False,
) -> List[tls.CipherSuite]:
    """Filters a list of cipher suites.

    Various match conditions can be specified. A cipher suite is filtered, if all
    given conditions match (AND-logic). All filtered cipher suites are returned in
    a list.

    Arguments:
        cs_list: A list of cipher suites to be filtered.
        key_algo: Optional match condition. If the key_algo (a combination of
            key_exch and key_auth, e.g. "DHE_RSA") of a cipher suite is in the
            given list, it is a match.
        key_exch: Optional match condition. If the key_exch (e.g. "ECDH") of a
            cipher suite is in the given list, it is a match.
        key_auth: Optional match condition. If the key_auth (e.g. "ECDSA") of a
            cipher suite is in the given list, it is a match.
        cipher_type: Optional match condition. If the cipher_type (e.g.
            "STREAM") of a cipher suite is in the list, it is a match.
        cipher_prim: Optional match condition. If the cipher_prim (e.g. "AES")
            of a cipher suite is in the given list, it is a match.
        cipher: Optional match condition. If the cipher (e.g. "AES_256_CCM_8")
            of a cipher suite is in the given list, it is a match.
        mac: Optional match condition. If the mac (e.g. "SHA384") of a cipher
            suite is in the give list, it is a match.
        version: Optional match condition. Cipher suites supported by the given
            TLS version are a match. This is rather rudimentary implemented:
            AEAD ciphers only for TLS1.2, specific ciphers only for TLS1.3.
        full_hs: Optional match condition. If the implementation supports a
            full handshake with a cipher suite, i.e. an encrypted connection
            can successfully established, it is a match. It means the key
            exchange, the symmetric cipher and the hash primitive are all
            supported.
        key_exchange_supported: Optional match condition. If the key exchange
            for a cipher suite is supported, it is a match. Note, that this
            does not mean that the symmetric cipher is supported as well.
        remove: An indication if the filtered cipher suites shall be removed
            from the original list of cipher suites. Defaults to False.

    Returns:
        The list of cipher suites that match all the given conditions.
    """

    filter_funcs = []
    if key_algo is not None:
        filter_funcs.append(lambda cs: cs.key_algo in key_algo)  # type: ignore

    if key_exch is not None:
        filter_funcs.append(
            lambda cs: getattr(cs.key_algo_struct, "key_ex_type", None)
            in key_exch  # type: ignore
        )

    if key_auth is not None:
        filter_funcs.append(
            lambda cs: getattr(cs.key_algo_struct, "key_auth", None)
            in key_auth  # type: ignore
        )

    if cipher_type is not None:
        filter_funcs.append(
            lambda cs: getattr(cs.cipher_struct, "c_type", None)
            in cipher_type  # type: ignore
        )

    if cipher_prim is not None:
        filter_funcs.append(
            lambda cs: getattr(cs.cipher_struct, "primitive", None)
            in cipher_prim  # type: ignore
        )

    if cipher is not None:
        filter_funcs.append(lambda cs: cs.cipher in cipher)  # type: ignore

    if mac is not None:
        filter_funcs.append(lambda cs: cs.mac in mac)  # type: ignore

    if version in [tls.Version.SSL30, tls.Version.TLS10, tls.Version.TLS11]:
        filter_funcs.append(
            lambda cs: cs.key_algo is not tls.KeyExchangeAlgorithm.TLS13_KEY_SHARE
            and (
                cs.cipher_struct is None
                or cs.cipher_struct.c_type is not tls.CipherType.AEAD
            )
        )

    if version is tls.Version.TLS12:
        filter_funcs.append(
            lambda cs: cs.key_algo is not tls.KeyExchangeAlgorithm.TLS13_KEY_SHARE
        )

    if version is tls.Version.TLS13:
        filter_funcs.append(
            lambda cs: cs.key_algo is tls.KeyExchangeAlgorithm.TLS13_KEY_SHARE
        )

    if full_hs is not None:
        filter_funcs.append(lambda cs: cs.full_hs is full_hs)

    if key_exchange_supported is not None:
        filter_funcs.append(
            lambda cs: cs.key_exchange_supported is key_exchange_supported
        )

    if tls.CipherSuite.TLS_FALLBACK_SCSV in cs_list:
        cs_list.remove(tls.CipherSuite.TLS_FALLBACK_SCSV)

    if tls.CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV in cs_list:
        cs_list.remove(tls.CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV)

    filtered = []
    for cs in cs_list:
        cs_details = get_cipher_suite_details(cs)
        if cs_details is None:
            continue

        match = True
        for filt_func in filter_funcs:
            if not filt_func(cs_details):
                match = False
                break

        if match:
            filtered.append(cs)

    if remove:
        for cs in filtered:
            cs_list.remove(cs)

    return filtered


def int_to_bytes(number: int) -> bytes:
    """Convert an integer to a byte string

    Arguments:
        number: the integer to convert

    Returns:
        just as many bytes as needed to represent the integer as an octet string
    """

    if number:
        return number.to_bytes((number.bit_length() + 7) // 8, "big")

    else:
        return b"\0"


def set_logging_format() -> None:
    """Initializes the format of log messages
    """

    logging.basicConfig(format="%(levelname)s: %(message)s")


def set_logging_level(level: str) -> None:
    """Sets the logging level

    Arguments:
        level: The logging level to use.
    """

    logging.getLogger().setLevel(level.upper())


class Log(object):
    """A class which implements relative time stamps.
    """

    start_time = None

    @classmethod
    def time(cls) -> str:
        """returns a time stamp relative to the time of the first call to this method.

        Returns:
            the time in seconds with 3 positions after the decimal point.
        """

        timestamp = time.time()
        if cls.start_time is None:
            cls.start_time = timestamp

        diff = timestamp - cls.start_time
        return f"Timestamp {diff:.3f}"


class Table(object):
    """Helper class for printing text in a table

    Attributes:
        indent: the indentation level, i.e. how many blanks shall be added on
            the left side of the table
        sep: the separator to print between adjacent columns within a row
    """

    def __init__(self, indent: int = 0, sep: str = ": ") -> None:
        self._indent = indent
        self._sep = sep
        self._nbr_columns = 0
        self._rows: List[List[Tuple[str, int]]] = []

    def row(self, *args: Union[Tuple[str, int], str]) -> None:
        """Register a row

        Arguments:
            args: a complete row, one string for each column
        """
        self._nbr_columns = max(self._nbr_columns, len(args))
        cols = [col if type(col) is tuple else (col, len(col)) for col in args]
        self._rows.append(cols)  # type: ignore

    def dump(self) -> None:
        """Print the table
        """

        if not self._nbr_columns:
            return

        cols = [1] * self._nbr_columns
        for row in self._rows:
            for idx, col in enumerate(row):
                cols[idx] = max(cols[idx], col[1])

        cols[-1] = 1
        for row in self._rows:
            print(" " * self._indent, end="")
            print(
                self._sep.join(
                    [
                        f"{col[0]:{cols[idx] + len(col[0]) - col[1]}}"
                        for idx, col in enumerate(row)
                    ]
                )
            )


def get_random_value() -> bytes:
    """Get a value suitable for a ClientHello or ServerHello

    Returns:
        32 bytes of almost random data
    """

    random = bytearray()
    random.extend(pdu.pack_uint32(int(time.time())))
    random.extend(os.urandom(28))
    return random


def exit_with_error(error: str) -> None:
    """Abort tlsmate with an error message

    Arguments:
        error: the error message to print on stderr
    """

    sys.stderr.write(f"Error: {error}\n")
    sys.stderr.flush()
    sys.exit(1)



# -*- coding: utf-8 -*-
"""Module scanning for the CCS-injection vulnerability

Refer to CVE-2014-0224

OpenSSL 1.0.1g and before was vulnerable.

ChangeCipherSpec is injected directly after receiving the ServerHelloDone. The server
then updated its connection state which means subsequent sent messages by the client
were regarded as protected records.
"""
# import basic stuff

# import own stuff
import tlsmate.msg as msg
import tlsmate.plugin as plg
import tlsmate.tls as tls

# import other stuff


class ScanCcsInjection(plg.Worker):
    name = "ccsinjection"
    descr = "scan for CCS injection vulnerability (CVE-2014-0224)"
    prio = 40

    def run(self):
        status = tls.ScanState.NA
        values = self.server_profile.get_profile_values(
            [tls.Version.TLS10, tls.Version.TLS11, tls.Version.TLS12], full_hs=True
        )
        if values.versions:
            status = tls.ScanState.UNDETERMINED
            self.client.init_profile(profile_values=values)
            with self.client.create_connection() as conn:
                conn.send(msg.ClientHello)
                conn.wait(msg.ServerHello)
                conn.wait(msg.Certificate, optional=True)
                conn.wait(msg.ServerKeyExchange, optional=True)
                conn.wait(msg.CertificateRequest, optional=True)
                conn.wait(msg.ServerHelloDone)

                # Normal handshake up to here. Now prematurely inject the CCS.
                conn.send(msg.ChangeCipherSpec)

                # And now send any other message. We use CCS again
                conn.send(msg.ChangeCipherSpec)

                try:
                    rec_msg = conn.wait(msg.Alert, timeout=2000, fail_on_timeout=False)
                    if isinstance(rec_msg, msg.Alert) and (
                        rec_msg.description is tls.AlertDescription.BAD_RECORD_MAC
                        or rec_msg.description is tls.AlertDescription.DECRYPTION_FAILED
                    ):
                        # vulnerable
                        status = tls.ScanState.TRUE

                    else:
                        status = tls.ScanState.FALSE

                except tls.TlsConnectionClosedError:
                    status = tls.ScanState.FALSE

        self.server_profile.allocate_vulnerabilities()
        self.server_profile.vulnerabilities.ccs_injection = status
